"""."""

import pytest

from kinematic_tracker.association.metric_giou_aligned import MetricGIoUAligned


def test_init(driver: MetricGIoUAligned) -> None:
    assert driver.aux_r.shape == (100,)


def test_num_z_ge_6() -> None:
    with pytest.raises(AssertionError):
        MetricGIoUAligned(10, 10, 5, (0, 1, 2, -3, -2, -1))



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.check import check_cov_zz


def test_undefined_cov_zz() -> None:
    check_cov_zz(np.full((5, 5), 1.0))
    with pytest.raises(ValueError):
        check_cov_zz(np.full((5, 5), np.nan))



"""The module for computing Generalized Intersection over Union (GIoU) association metric.

This module provides functionality for computing the similarity score between detection reports
and Kalman filter states (targets) using the Generalized Intersection over Union (GIoU) metric.
The metric is computed between the 3D bounding boxes which are oriented by yaw angle.

Classes:
    MetricGIoU: GIoU association metrics between detections and Kalman filters.
"""

from typing import Sequence

import numpy as np

from kinematic_tracker.geometry.giou_yaw_aux import GiouYawAux
from kinematic_tracker.geometry.rectangle_buffers import RectangleBuffers
from kinematic_tracker.types import FMT, FT

from .metric_driver_base import MetricDriverBase


class MetricGIoUYaw(MetricDriverBase):
    """Computes GIoU association metric between detections and Kalman filter states.

    This class inherits from MetricDriverBase and implements GIoU-based metric computation
    for data association between detection reports and Kalman filter states. The computation
    is performed between cuboids (3D bounding boxes) rotated by yaw angle (around z axis
    by default).

    Args:
        num_reports_max: Maximum number of detection reports to handle.
        num_targets_max: Maximum number of targets (Kalman filters) to handle.
        num_z: Dimension of the measurement vector.
    """

    def __init__(
        self,
        num_reports_max: int,
        num_targets_max: int,
        num_z: int,
        ind_pos_w_rz_size: Sequence[int],
    ) -> None:
        """Initialize the MetricGIoUAligned instance.

        Args:
            num_reports_max: maximum number of detection reports to handle.
            num_targets_max: maximum number of targets (Kalman filters) to handle.
            num_z: number of variables in the measurement vector.
            ind_pos_w_rz_size: location of variables in the measurement vector.
                          By default, the indices are 0,1,2,3,6,7,8,9, i.e.
                          the center of the cuboid is taken as first three variables,
                          the scalar part of the quaternion follows the center,
                          the z-projection of the quaternion follows the scalar part,
                          and sizes (dimensions) of the cuboids are taken from
                          the last three variables.
        """
        assert num_z > 7, 'Number of variables should be 8 or greater.'
        super().__init__(num_reports_max, num_targets_max, num_z)
        self.ind_pos_w_rz_size = np.array(ind_pos_w_rz_size, dtype=int)
        assert self.ind_pos_w_rz_size.ndim == 1, 'Indices should be one-dimensional.'
        assert self.ind_pos_w_rz_size.size == 8, 'There should be 8 fancy indices.'
        self.report_buf = RectangleBuffers(num_reports_max)
        self.target_buf = RectangleBuffers(num_targets_max)
        self.xx_tz = np.zeros((num_targets_max, 10))
        self.xx_rz = np.zeros((num_reports_max, 10))
        self.xx_idx = np.array((0, 1, 2, 3, 6, 7, 8, 9), int)
        self.volumes_t = np.zeros(num_targets_max)
        self.aux = GiouYawAux(num_targets_max)

    def set_det_rz(self, det_rz: FMT) -> None:
        num_r = det_rz.shape[0]
        self.xx_rz[:num_r, self.xx_idx] = det_rz[:, self.ind_pos_w_rz_size]
        self.report_buf.set_nu_scenes(self.xx_rz[:num_r])

    def set_target_tz(self, filters: FT) -> None:
        for t, kf in enumerate(filters):
            np.dot(kf.measurementMatrix, kf.statePre[:, 0], out=self.vec_z[:, 0])
            self.xx_tz[t, self.xx_idx] = self.vec_z[self.ind_pos_w_rz_size, 0]
        num_t = len(filters)
        self.target_buf.set_nu_scenes(self.xx_tz[:num_t])
        self.volumes_t[:num_t] = self.xx_tz[:num_t, 7:].prod(axis=1)

    def compute_metric(self, det_rz: FMT, filters: FT) -> FMT:
        """Compute GIoU association metric between detections and Kalman filter states.

        Args:
            det_rz: Detection measurements, either as a 2D numpy array.
            filters: Sequence of OpenCV KalmanFilter objects representing tracked targets.

        Returns:
            np.ndarray: A matrix of GIoU scores between each detection-target pair,
                with shape (num_detections, num_targets).
        """
        self.set_det_rz(det_rz)
        self.set_target_tz(filters)
        num_r = det_rz.shape[0]
        num_t = len(filters)
        metric_rt = self.get_rect_chunk(num_r, num_t)
        trk_tz = self.xx_tz[:num_t]
        trk_poly_t = self.target_buf.mid_polygons_n[:num_t]
        volumes_t = self.volumes_t[:num_t]
        self.aux.set_second_4_vertices(self.target_buf.rect_coordinates_n4c[:num_t, :, :2])

        for r_num in range(num_r):
            det_z = self.xx_rz[r_num]
            det_poly = self.report_buf.mid_polygons_n[r_num]
            rect_4c = self.report_buf.rect_coordinates_n4c[r_num, :, :2]
            self.aux.set_first_4_vertices(rect_4c, num_t)
            self.aux.compute(det_z, det_poly, trk_tz, trk_poly_t, volumes_t, metric_rt[r_num])
        return metric_rt



"""."""

from kinematic_tracker import NdKkfTracker


def test_nd_kkf_tracker_repr(tracker: NdKkfTracker) -> None:
    """."""
    ref = """NdKkfTracker(
    Association(greedy mahalanobis threshold 0.56 mah_pre_factor 1.32)
    num_misses_max 3
    num_det_min 2)"""
    assert repr(tracker) == ref



"""."""

import cv2
import numpy as np
import pytest

from kinematic_tracker.association.metric_giou_yaw import MetricGIoUYaw


def test_set_target_tz(driver: MetricGIoUYaw, filters: list[cv2.KalmanFilter]) -> None:
    driver.set_target_tz(filters)
    assert driver.target_buf.num_rect == 3
    ref = [
        [1.0, 2.0, 3.0, 4.0, 0.0, 0.0, 7.0, 8.0, 9.0, 10.0],
        [11.0, 12.0, 13.0, 14.0, 0.0, 0.0, 17.0, 18.0, 19.0, 20.0],
        [21.0, 22.0, 23.0, 24.0, 0.0, 0.0, 27.0, 28.0, 29.0, 30.0],
    ]
    assert driver.xx_tz[:3] == pytest.approx(np.array(ref))
    assert driver.volumes_t[:3] == pytest.approx([720.0, 6840.0, 24360.0])



"""."""

import numpy as np
import shapely

from scipy.spatial.transform import Rotation
from shapely import Polygon

from kinematic_tracker.types import FMT


SPT = np.ndarray[tuple[int], np.dtype[np.object_]]
MID_VERTICES_C4 = ((0.5, 0.5, -0.5, -0.5), (-0.5, 0.5, 0.5, -0.5), (0.0, 0.0, 0.0, 0.0))


class RectangleBuffers:
    def __init__(self, num_max: int) -> None:
        self.unit_vertices_14c = np.array(MID_VERTICES_C4).T.reshape(1, 4, 3)
        self.num_max = num_max
        self.mid_vertices_n4c = np.zeros((num_max, 4, 3))
        self.rect_coordinates_n4c = np.zeros((num_max, 4, 3))
        self.mid_polygons_n: SPT = np.zeros(num_max, dtype=Polygon)
        self.rotations = Rotation.identity(shape=(num_max, 1))
        self.num_rect = 0

    def set_nu_scenes(self, vecs_nz: FMT) -> None:
        assert vecs_nz.ndim == 2 and vecs_nz.shape[1] == 10
        self.num_rect = n = vecs_nz.shape[0]
        if n == 0:
            return
        sizes_ptr = vecs_nz[:, 7:].reshape(n, 1, 3)
        np.multiply(sizes_ptr, self.unit_vertices_14c, out=self.mid_vertices_n4c[:n])
        rotations_ptr = self.rotations[:n]
        rotations_ptr[:] = Rotation.from_quat(vecs_nz[:, 3:7].reshape(n, 1, 4), scalar_first=True)
        self.rect_coordinates_n4c[:n] = rotations_ptr.apply(self.mid_vertices_n4c[:n])
        self.rect_coordinates_n4c[:n] += vecs_nz[:, :3].reshape(n, 1, 3)
        self._mid_polygons()

    def _mid_polygons(self) -> None:
        n = self.num_rect
        shapely.polygons(self.rect_coordinates_n4c[:n, :, :2], out=self.mid_polygons_n[:n])



import numpy as np
import pytest

from kinematic_tracker.tracker.check import check_ids_r


def test_when_sizes_are_wrong() -> None:
    ids_r = np.linspace(0, 3, 4, dtype=int)
    with pytest.raises(ValueError) as err_size:
        check_ids_r(ids_r, 5)
    assert err_size.value.args[0] == 'I expect 5 ids, but I see 4.'

    with pytest.raises(ValueError) as err_dim:
        check_ids_r(ids_r.reshape(1, 4), 4)
    assert err_dim.value.args[0] == 'I expect one-dimensional vector, but I see 2.'


def test_when_data_type_is_wrong() -> None:
    with pytest.raises(TypeError) as err_dtype:
        check_ids_r(np.ones(4), 4)
    assert err_dtype.value.args[0] == 'I expect integer ids, but I see float64.'


def test_when_type_is_wrong() -> None:
    with pytest.raises(TypeError) as err_dtype:
        check_ids_r([0], 4)
    assert err_dtype.value.args[0] == "I expect NumPy array, but I see <class 'list'>."


def test_everything_ok() -> None:
    ids_r = np.linspace(0, 3, 4, dtype=int)
    check_ids_r(ids_r, 4)
    assert True



"""."""

import pytest

from kinematic_tracker.association.metric_giou_yaw import MetricGIoUYaw


def test_init(driver: MetricGIoUYaw) -> None:
    assert driver.xx_rz.shape == (15, 10)
    assert driver.xx_tz.shape == (17, 10)


def test_num_z_ge_8() -> None:
    with pytest.raises(AssertionError):
        MetricGIoUYaw(10, 10, 7, (0, 1, 2, 3, 6, -3, -2, -1))


def test_fancy_indices() -> None:
    with pytest.raises(AssertionError):
        MetricGIoUYaw(10, 10, 8, (0, 2, 3, 6, -3, -2, -1))

    with pytest.raises(AssertionError):
        MetricGIoUYaw(10, 10, 8, [(0, 1, 2, 3, 6, -3, -2, -1)])



"""."""

from kinematic_tracker.association.association import Association
from kinematic_tracker.association.association_metric import AssociationMetric


def test_init() -> None:
    """."""
    association = Association(12, 6)
    assert association.metric == AssociationMetric.MAHALANOBIS



"""."""

import numpy as np
import pytest

from kinematic_tracker.geometry.giou_yaw_aux import GiouYawAux
from kinematic_tracker.types import F3T


def test_comp_heights(aux: GiouYawAux, vecs_289: F3T) -> None:
    cuboids = np.linspace(1, 30, 30).reshape(3, 10)
    cuboid = np.linspace(1, 10, 10)
    intersections, unions = aux.comp_heights(cuboid, cuboids)
    assert intersections == pytest.approx([10.0, 5.0, 0.0])
    assert unions == pytest.approx([10.0, 25.0, 40.0])



"""."""

import copy

from typing import Callable

import numpy as np

from kinematic_tracker.types import FMT, IVT


class ScoreCopy:
    def __init__(self, num_cls: int) -> None:
        self.num_cls = num_cls
        self.score_crt = np.empty(num_cls, dtype=object)
        self.creation_ids_ct = np.empty(num_cls, dtype=object)

    def set(self, cls: int, creation_ids: Callable[[int], IVT], score_rt: FMT) -> None:
        self.score_crt[cls] = score_rt.copy()
        self.creation_ids_ct[cls] = creation_ids(cls)

    def deep_copy(self) -> 'ScoreCopy':
        return copy.deepcopy(self)


class StubScoreCopy:
    def __repr__(self) -> str:
        return 'Stub for the ScoreCopy. NdKkfTracker(..., return_score=True) maybe?'

    def set(self, _cls: int, _creation_ids: Callable[[int], IVT], _score_rt: FMT) -> None: ...

    def deep_copy(self) -> 'StubScoreCopy':
        return self


def get_score_copy(return_score: bool, num_cls: int) -> ScoreCopy | StubScoreCopy:
    if return_score:
        return ScoreCopy(num_cls)
    return StubScoreCopy()



"""."""

from kinematic_tracker import NdKkfTracker


def test_single_class(tracker1: NdKkfTracker) -> None:
    tracker1.clear_tracks()
    assert tracker1.tracks_c[0] == []


def test_two_classes(tc2: NdKkfTracker) -> None:
    tc2.clear_tracks()
    assert tc2.tracks_c[0] == []
    assert tc2.tracks_c[1] == []



"""."""

from kinematic_tracker.association.association import Association
from kinematic_tracker.association.association_metric import AssociationMetric
from kinematic_tracker.association.metric_giou_aligned import MetricGIoUAligned
from kinematic_tracker.association.metric_giou_yaw import MetricGIoUYaw
from kinematic_tracker.association.metric_mahalanobis import MetricMahalanobis
from kinematic_tracker.association.metric_mahalanobis_size_mod import MetricMahalanobisSizeMod


def test_get_num_z_6(association: Association) -> None:
    """."""
    association.metric = AssociationMetric.MAHALANOBIS
    assert isinstance(association.get_metric_driver(), MetricMahalanobis)

    association.metric = AssociationMetric.GIOU_AXES_ALIGNED
    assert isinstance(association.get_metric_driver(), MetricGIoUAligned)

    association.metric = AssociationMetric.MAHALANOBIS_SIZE_MODULATED
    assert isinstance(association.get_metric_driver(), MetricMahalanobisSizeMod)


def test_giou_yaw() -> None:
    association = Association(16, 10)
    association.metric = AssociationMetric.GIOU_YAW
    assert isinstance(association.get_metric_driver(), MetricGIoUYaw)



"""Types of association metrics (similarity scores)."""

from enum import Enum


class AssociationMetric(Enum):
    GIOU_AXES_ALIGNED = 'giou-axes-aligned'
    GIOU_YAW = 'giou-yaw'
    MAHALANOBIS = 'mahalanobis'
    MAHALANOBIS_SIZE_MODULATED = 'mahalanobis-size-modulated'



"""."""

import numpy as np

from kinematic_tracker import NdKkfTracker
from kinematic_tracker.tracker.targets import upd_for_loose_tracks


def test_upd_for_loose_tracks1(tracker1: NdKkfTracker) -> None:
    upd_for_loose_tracks(tracker1.tracks_c[0], np.zeros(1, int), 1, tracker1.num_misses_max)
    assert len(tracker1.tracks_c[0]) == 1
    assert tracker1.tracks_c[0][0].num_miss == 0



"""Commonly used types."""

from typing import Sequence

import cv2
import numpy as np


F3T = np.ndarray[tuple[int, int, int], np.dtype[np.float64]]
FMT = np.ndarray[tuple[int, int], np.dtype[np.float64]]
FVT = np.ndarray[tuple[int], np.dtype[np.float64]]
IVT = np.ndarray[tuple[int], np.dtype[np.int64]]
DT = Sequence[FVT] | FMT | Sequence[Sequence[float]]
FT = Sequence[cv2.KalmanFilter]



"""."""

import pytest

from kinematic_tracker import NdKkfTracker
from kinematic_tracker.tracker.track import NdKkfTrack


def test_step1(tc2: NdKkfTracker) -> None:
    assert len(tc2.tracks_c) == 2
    assert len(tc2.tracks_c[0]) == 1
    assert len(tc2.tracks_c[1]) == 2

    track00: NdKkfTrack = tc2.tracks_c[0][0]
    assert track00.creation_id == 0
    assert track00.kkf.kalman_filter.statePost[:, 0] == pytest.approx(
        [1, 0, 0, 2, 0, 0, 3, 0, 0, 4, 5, 6]
    )

    track10: NdKkfTrack = tc2.tracks_c[1][0]
    assert track10.creation_id == 1
    ref10 = 7, 0, 0, 8, 0, 0, 9, 0, 0, 10, 11, 12
    assert track10.kkf.kalman_filter.statePost[:, 0] == pytest.approx(ref10)

    track11: NdKkfTrack = tc2.tracks_c[1][1]
    assert track11.creation_id == 2
    ref11 = 13, 0, 0, 14, 0, 0, 15, 0, 0, 16, 17, 18
    assert track11.kkf.kalman_filter.statePost[:, 0] == pytest.approx(ref11)



"""."""

import numpy as np
import pytest

from kinematic_tracker.association.metric_driver_base import FT, MetricDriverBase


def test_compute(driver: MetricDriverBase, filters: FT) -> None:
    driver = MetricDriverBase(4, 5, 6)
    vxx_z = [vec_z.copy() for vec_z in driver.gen_vec_z(filters)]
    assert len(vxx_z) == 3
    assert vxx_z[0] == pytest.approx([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
    assert vxx_z[1] == pytest.approx([2.0, 3.0, 4.0, 5.0, 6.0, 7.0])
    assert vxx_z[2] == pytest.approx([3.0, 4.0, 5.0, 6.0, 7.0, 8.0])


def test_share_buffer(driver: MetricDriverBase, filters: FT) -> None:
    ls = list(driver.gen_vec_z(filters))
    assert len(ls) == 3
    assert np.shares_memory(ls[0], driver.vec_z)
    assert np.shares_memory(ls[1], driver.vec_z)
    assert np.shares_memory(ls[2], driver.vec_z)



"""."""

from kinematic_tracker import NdKkfTracker


def test_reset1(tracker1: NdKkfTracker) -> None:
    tracker1.reset()
    assert tracker1.tracks_c.shape == (1,)
    assert len(tracker1.tracks_c[0]) == 0
    assert tracker1.last_ts_ns == 0


def test_reset2(tc2: NdKkfTracker) -> None:
    tc2.reset()
    assert tc2.tracks_c.shape == (2,)
    assert len(tc2.tracks_c[0]) == 0
    assert len(tc2.tracks_c[1]) == 0



"""."""

from typing import Any, Sequence

from .free_id_builtin import get_free_id_builtin


def get_free_id(tracks: Sequence[Any]) -> int:
    return get_free_id_builtin([track.id for track in tracks])



import numpy as np
import shapely

from shapely import Polygon

from kinematic_tracker.geometry.rectangle_buffers import SPT
from kinematic_tracker.types import F3T, FMT, FVT


class GiouYawAux:
    def __init__(self, num_max: int) -> None:
        self.merged_p8c = np.zeros((num_max, 8, 2))
        self.multi_points_p = np.zeros(num_max, object)
        self.convex_hulls_p = np.zeros(num_max, object)
        self.hull_areas_p = np.zeros(num_max)
        self.intersection_areas_p = np.zeros(num_max)
        self.intersection_heights_p = np.zeros(num_max)
        self.union_heights_p = np.zeros(num_max)
        self.intersection_volumes_p = np.zeros(num_max)
        self.union_volumes_p = np.zeros(num_max)
        self.convex_volumes_p = np.zeros(num_max)
        self.za_m_zb_p = np.zeros(num_max)
        self.averaged_heights_p = np.zeros(num_max)
        self.stacked_4p = np.zeros((4, num_max))

    def comp_heights(self, cuboid: FVT, cuboids: FMT) -> tuple[FVT, FVT]:
        n = cuboids.shape[0]
        intersections = self.intersection_heights_p[:n]
        unions = self.union_heights_p[:n]
        za_m_zb = self.za_m_zb_p[:n]
        za_m_zb[:] = cuboid[2]
        za_m_zb -= cuboids[:, 2]
        stacked_4p = self.stacked_4p[:, :n]
        stacked_4p[3] = cuboids[:, -1]
        averaged_heights = self.averaged_heights_p[:n]
        averaged_heights[:] = cuboid[-1]
        averaged_heights += stacked_4p[3, :]
        averaged_heights /= 2.0
        stacked_4p[0] = averaged_heights
        stacked_4p[0] += za_m_zb
        stacked_4p[1] = averaged_heights
        stacked_4p[1] -= za_m_zb
        stacked_4p[2] = cuboid[-1]
        np.amin(stacked_4p, axis=0, out=intersections)
        np.maximum(intersections, 0, out=intersections)
        np.amax(stacked_4p, axis=0, out=unions)
        return intersections, unions

    def set_first_4_vertices(self, rect_4c: FMT, n: int) -> None:
        self.merged_p8c[:n, :4] = rect_4c

    def set_second_4_vertices(self, rect_t4c: F3T) -> None:
        n = rect_t4c.shape[0]
        self.merged_p8c[:n, 4:] = rect_t4c

    def comp_convex_hull_areas(self, n: int) -> FMT:
        shapely.multipoints(self.merged_p8c[:n], out=self.multi_points_p[:n])
        shapely.convex_hull(self.multi_points_p[:n], out=self.convex_hulls_p[:n])
        shapely.area(self.convex_hulls_p[:n], out=self.hull_areas_p[:n])
        return self.hull_areas_p[:n]

    def compute(
        self, cuboid: FVT, poly: Polygon, cuboids: FMT, polygons: SPT, volumes: FVT, out: FVT
    ) -> None:
        n = polygons.shape[0]
        shapely.area(shapely.intersection(poly, polygons), out=self.intersection_areas_p[:n])
        ch_areas = self.comp_convex_hull_areas(n)
        intersection_heights, union_heights = self.comp_heights(cuboid, cuboids)

        intersection_volumes = self.intersection_volumes_p[:n]
        intersection_volumes[:] = self.intersection_areas_p[:n]
        intersection_volumes *= intersection_heights

        union_volumes = self.union_volumes_p[:n]
        union_volumes[:] = cuboid[-3:].prod()
        union_volumes += volumes[:]
        union_volumes -= intersection_volumes
        convex_vol = self.convex_volumes_p[:n]
        np.multiply(ch_areas, union_heights, out=convex_vol)

        tmp_vol = union_volumes

        out[:] = intersection_volumes
        out /= tmp_vol
        tmp_vol /= convex_vol
        out += tmp_vol
        out /= 2.0



import cv2
import numpy as np
import pytest

from kinematic_tracker.association.metric_giou_yaw import MetricGIoUYaw
from kinematic_tracker.types import FMT


def test_normal(driver: MetricGIoUYaw, filters: list[cv2.KalmanFilter], det_rz: FMT) -> None:
    metric_rt = driver.compute_metric(det_rz, filters)
    ref = [
        [1.0, 0.3185216546083782, 0.26513977158265706],
        [0.3185216546083782, 1.0, 0.4392318438129806],
    ]
    assert metric_rt == pytest.approx(np.array(ref))



"""."""

import pytest

from kinematic_tracker.association.metric_mean_dim_giou import MetricMeanDimGIoU


@pytest.fixture
def driver() -> MetricMeanDimGIoU:
    return MetricMeanDimGIoU(100, 500, 6, (0, 1, 2, -3, -2, -1))



"""."""

from kinematic_tracker.tracker.score_copy import ScoreCopy


def test_init() -> None:
    sp = ScoreCopy(1)
    assert sp.score_crt[0] is None
    assert sp.creation_ids_ct[0] is None
    assert sp.creation_ids_ct.ndim == 1
    assert sp.creation_ids_ct.shape[0] == 1
    assert sp.score_crt.ndim == 1
    assert sp.score_crt.shape[0] == 1



"""."""

import pytest

from kinematic_tracker.association.metric_mahalanobis_size_mod import MetricMahalanobisSizeMod


@pytest.fixture
def driver() -> MetricMahalanobisSizeMod:
    return MetricMahalanobisSizeMod(100, 500, 1.23, 12, 6, (0, 1, 2, -3, -2, -1))



"""."""

import numpy as np
import pytest

from kinematic_tracker.association.metric_mean_dim_giou import MetricMeanDimGIoU


def test_fill_target_len_min_max(driver: MetricMeanDimGIoU) -> None:
    vec_z = np.linspace(1.0, 6.0, num=6)
    driver.fill_target_len_min_max(vec_z, driver.len_s, driver.min_s, driver.max_s)
    assert driver.len_s == pytest.approx([4.0, 5.0, 6.0])
    assert driver.min_s == pytest.approx([-1.0, -0.5, 0.0])
    assert driver.max_s == pytest.approx([3.0, 4.5, 6.0])



"""."""

import numpy as np
import pytest

from kinematic_tracker import NdKkfTracker


def test_with_measurement_orders() -> None:
    tracker = NdKkfTracker([3, 1], [3, 3], [2, 1])
    assert tracker.cov_zz.shape == (9, 9)
    cov_zz_copy = tracker.cov_zz.copy()
    np.fill_diagonal(cov_zz_copy, 0.0)
    assert cov_zz_copy == pytest.approx(0.0)
    assert np.all(np.isnan(tracker.cov_zz.diagonal()))
    assert tracker.gen_xz.orders_z == pytest.approx([2, 1])
    assert tracker.gen_xz.gen_x.num_x == 12
    assert tracker.gen_xz.num_z == 9



"""."""

import numpy as np

from kinematic_tracker.geometry.giou_yaw_aux import GiouYawAux


def test_set_first_4_vertices(aux: GiouYawAux) -> None:
    rect_4c = np.linspace(1, 8, 8).reshape(4, 2)
    aux.set_first_4_vertices(rect_4c, 6)
    ref8 = [
        [1.0, 2.0],
        [3.0, 4.0],
        [5.0, 6.0],
        [7.0, 8.0],
        [0.0, 0.0],
        [0.0, 0.0],
        [0.0, 0.0],
        [0.0, 0.0],
    ]
    assert np.allclose(aux.merged_p8c[:6], np.array(ref8))



"""."""

import numpy as np

from kinematic_tracker.association.metric_giou_yaw import FT, MetricGIoUYaw


def test_empty_det(filters: FT, driver: MetricGIoUYaw) -> None:
    metric = driver.compute_metric(np.empty((0, 10)), filters)
    assert metric.shape == (0, 3)



"""."""

from kinematic_tracker.geometry.rectangle_buffers import RectangleBuffers


def test_init(rb5: RectangleBuffers) -> None:
    assert rb5.rect_coordinates_n4c.shape == (5, 4, 3)
    assert rb5.num_rect == 0



"""."""

from kinematic_tracker.association.metric_giou_aligned import FT, MetricGIoUAligned


def test_empty_det(filters: FT, driver: MetricGIoUAligned) -> None:
    metric = driver.compute_metric([], filters)
    assert metric.shape == (0, 3)



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.score_copy import StubScoreCopy


@pytest.fixture
def ssp() -> StubScoreCopy:
    return StubScoreCopy()


def test_init(ssp: StubScoreCopy) -> None:
    assert vars(ssp) == {}


def test_set(ssp: StubScoreCopy) -> None:
    ssp.set(1, lambda c: np.zeros(1, dtype=int), np.ones((1, 1)))
    assert vars(ssp) == {}


def test_deep_copy(ssp: StubScoreCopy) -> None:
    assert ssp.deep_copy() == ssp


def test_repr(ssp: StubScoreCopy) -> None:
    assert repr(ssp) == 'Stub for the ScoreCopy. NdKkfTracker(..., return_score=True) maybe?'



"""."""

import numpy as np
import pytest

from kinematic_tracker.association.metric_giou_aligned import MetricGIoUAligned
from kinematic_tracker.types import DT


def test_comp_report_aux_normal(det_rz: DT, driver: MetricGIoUAligned) -> None:
    driver.comp_report_aux(det_rz)
    ref0 = [[-1.0, -0.5, 0.0], [3.0, 4.5, 6.0]]
    assert driver.aux_r[0].corners == pytest.approx(np.array(ref0))
    ref1 = [[-0.5, 0.0, 0.5], [4.5, 6.0, 7.5]]
    assert driver.aux_r[1].corners == pytest.approx(np.array(ref1))
    assert driver.aux_r[2].corners == pytest.approx(0.0)



"""."""

from kinematic_tracker import NdKkfTracker


def test_get_filters(tracker1: NdKkfTracker) -> None:
    filters = tracker1.get_kalman_filters(0)
    assert len(filters) == 1


def test_multi_class(tc2: NdKkfTracker) -> None:
    filters0 = tc2.get_kalman_filters(0)
    assert len(filters0) == 1

    filters1 = tc2.get_kalman_filters(1)
    assert len(filters1) == 2



"""."""

import numpy as np
import pytest

from kinematic_tracker.association.metric_mean_dim_giou import MetricMeanDimGIoU
from kinematic_tracker.types import DT


def test_fill_det_len_min_max(det_rz: DT, driver: MetricMeanDimGIoU) -> None:
    driver.fill_det_len_min_max(det_rz, driver.len_rs, driver.min_rs, driver.max_rs)
    assert driver.len_rs[:2] == pytest.approx(np.array([[4.0, 5.0, 6.0], [5.0, 6.0, 7.0]]))
    assert driver.len_rs[2:] == pytest.approx(0.0)
    assert driver.min_rs[:2] == pytest.approx(np.array([[-1.0, -0.5, 0.0], [-0.5, 0.0, 0.5]]))
    assert driver.min_rs[2:] == pytest.approx(0.0)
    assert driver.max_rs[2:] == pytest.approx(0.0)
    assert driver.max_rs[:2] == pytest.approx(np.array([[3.0, 4.5, 6.0], [4.5, 6.0, 7.5]]))



import numpy as np

from kinematic_tracker.types import FMT, IVT


TYPE_ERR_MSG = 'Time stamp (nanoseconds) should be an integer or np.int64, but got '
SIZE_ERR_MSG = 'Wrong size of the measurement vector '


def check_det_rz(det_rz: FMT, num_z: int) -> None:
    if not isinstance(det_rz, np.ndarray):
        raise TypeError(f'Expected (two-dimensional) NumPy array, got {type(det_rz)}')
    if det_rz.ndim != 2:
        raise ValueError(f'Expected two-dimensional NumPy array, got {det_rz.ndim}')
    if det_rz.shape[1] != num_z:
        raise ValueError(f'{SIZE_ERR_MSG}{det_rz.shape[1]}. Expected {num_z}')


def check_stamp_type(stamp: int | np.int64) -> None:
    if not isinstance(stamp, int) and not isinstance(stamp, np.int64):
        raise TypeError(f'{TYPE_ERR_MSG}{type(stamp)}')


def check_cov_zz(cov_zz: FMT) -> None:
    """Ensure the measurement covariance is fully defined.

    Raises:
        ValueError: if the measurement covariance contains NaN values.
    """
    if np.isnan(cov_zz.sum()):
        raise ValueError('Not all elements of measurement covariance are defined.')


def check_cls_r(cls_r: IVT, num_reports: int, num_cls: int) -> None:
    """Ensure classes matches the reports and are within bounds."""
    if cls_r.shape[0] != num_reports:
        raise ValueError(f'Number of reports should be {num_reports}, but got {cls_r.shape[0]}.')
    if num_reports < 1:
        return
    if cls_r.min() < 0:
        raise ValueError(f'Class IDs should be greater or equal zero, but I see {cls_r.min()}.')
    if cls_r.max() >= num_cls:
        raise ValueError(f'Class IDs should be less than {num_cls}, but I see {cls_r.max()}.')


def check_ids_r(ids_r: IVT, num_reports: int) -> None:
    """Ensure the annotation ids are of correct type and size.

    Raises:
        ValueError: if the dimensionality of the ids vector is incorrect.
        TypeError: if the vector of IDs is not integer-valued.
    """
    if not isinstance(ids_r, np.ndarray):
        raise TypeError(f'I expect NumPy array, but I see {type(ids_r)}.')

    if ids_r.ndim == 1 and ids_r.shape[0] == num_reports and np.issubdtype(ids_r.dtype, np.integer):
        return
    if ids_r.ndim != 1:
        raise ValueError(f'I expect one-dimensional vector, but I see {ids_r.ndim}.')
    elif ids_r.shape[0] != num_reports:
        raise ValueError(f'I expect {num_reports} ids, but I see {ids_r.shape[0]}.')
    elif not np.issubdtype(ids_r.dtype, np.integer):
        raise TypeError(f'I expect integer ids, but I see {ids_r.dtype}.')



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.check import check_det_rz


def test_is_np() -> None:
    with pytest.raises(TypeError) as err:
        check_det_rz([[]], 6)
    assert err.value.args[0] == "Expected (two-dimensional) NumPy array, got <class 'list'>"


def test_is_2d() -> None:
    with pytest.raises(ValueError) as err:
        check_det_rz(np.zeros(1), 6)

    assert err.value.args[0] == 'Expected two-dimensional NumPy array, got 1'


def test_is_num_z_ok() -> None:
    with pytest.raises(ValueError) as err:
        check_det_rz(np.zeros((1, 5)), 6)

    assert err.value.args[0] == 'Wrong size of the measurement vector 5. Expected 6'



"""."""

import numpy as np
import pytest

from kinematic_tracker import NdKkfTracker
from kinematic_tracker.tracker.score_copy import StubScoreCopy
from kinematic_tracker.tracker.tracker import UNKNOWN_REPORT_ID


def test_with_report_ids(tracker1: NdKkfTracker) -> None:
    """."""
    assert tracker1.last_ts_ns == 1234_000_000
    assert tracker1.pre.last_dt == pytest.approx(1.234)
    assert len(tracker1.tracks_c[0]) == 1
    track = tracker1.tracks_c[0][0]
    assert track.ann_id == 123
    assert track.upd_id == 123
    assert track.num_miss == 0
    assert track.num_det == 0
    vec_x_ref = [1, 0, 0, 2, 0, 0, 3, 0, 0, 4, 5, 6]
    assert track.kkf.kalman_filter.statePost[:, 0] == pytest.approx(vec_x_ref)


def test_without_report_ids(tracker: NdKkfTracker) -> None:
    """."""
    tracker.set_measurement_cov(0.01 * np.eye(6))
    rv = tracker.advance(1000_000_000, np.linspace(1.0, 6.0, num=6).reshape(1, 6))
    assert isinstance(rv, StubScoreCopy)
    assert len(tracker.tracks_c[0]) == 1
    track = tracker.tracks_c[0][0]
    assert track.ann_id == UNKNOWN_REPORT_ID
    assert track.upd_id == UNKNOWN_REPORT_ID



"""."""

from kinematic_tracker.geometry.giou_yaw_aux import GiouYawAux


def test_init(aux: GiouYawAux) -> None:
    assert aux.merged_p8c.shape == (123, 8, 2)
    assert aux.stacked_4p.shape == (4, 123)
    assert aux.union_heights_p.shape == (123,)
    assert aux.union_volumes_p.shape == (123,)
    assert aux.intersection_volumes_p.shape == (123,)
    assert aux.intersection_areas_p.shape == (123,)
    assert aux.intersection_heights_p.shape == (123,)
    assert aux.hull_areas_p.shape == (123,)
    assert aux.convex_hulls_p.shape == (123,)
    assert aux.multi_points_p.shape == (123,)



"""."""

import numpy as np
import pytest

from kinematic_tracker import NdKkfTracker


def test_it(tracker: NdKkfTracker) -> None:
    with pytest.raises(TypeError) as err:
        tracker.advance(123.0, np.zeros((1, 6)))

    ref = "Time stamp (nanoseconds) should be an integer or np.int64, but got <class 'float'>"
    assert err.value.args[0] == ref



"""."""

import numpy as np

from kinematic_tracker.tracker.targets import upd_for_loose_tracks


def test_nd_kkf_tracker_upd_for_loose_tracks() -> None:
    tracks = []
    upd_for_loose_tracks(tracks, np.empty(0, int), 0, 3)
    assert tracks == []



from typing import Sequence

from kinematic_tracker.core.loose_items import get_loose_indices
from kinematic_tracker.types import FMT, IVT

from .track import NdKkfTrack


def upd_for_loose_tracks(
    tracks: list[NdKkfTrack], targets: IVT, num_tracks_before: int, num_misses_max: int
) -> None:
    """Update the list of tracks for loose tracks (i.e., tracks without association).

    Args:
        tracks: list of tracks to be updated.
        targets: Indices of associated tracks.
        num_tracks_before: Number of tracks before association.
        num_misses_max: Number of misses to tolerate before removal of a track.
    """
    tracks_wo_correspondence = get_loose_indices(targets, num_tracks_before)
    for t in tracks_wo_correspondence:
        tracks[t].bump_num_misses()

    to_remove = [t for t in tracks if t.num_miss > num_misses_max]
    for t in to_remove:
        tracks.remove(t)


# fmt: off
def correct_associated(tracks: Sequence[NdKkfTrack], reports: IVT, targets: IVT, metric_rt: FMT, det_rz: FMT, ids_r: IVT) -> None:
# fmt: on
    """Correct tracks based on the given association matches.

    Args:
        tracks: Tracks to be eventually corrected.
        reports: Indices of associated detection reports.
        targets: Indices of associated tracked targets.
        metric_rt: Metric values for report-target pairs.
        det_rz: Detection vectors to be used eventually for correction step.
        ids_r: Detection IDs to be used eventually for assigning the update IDs.
    """
    for r, t in zip(reports, targets):
        track = tracks[t]
        track.kkf.correct(det_rz[r])
        track.bump_num_detections(float(metric_rt[r, t]), int(ids_r[r]))



"""."""

import numpy as np

from kinematic_tracker import NdKkfTracker
from kinematic_tracker.tracker.targets import upd_for_loose_tracks


def test_remove_track(tracker1: NdKkfTracker) -> None:
    """."""
    no_targets = np.empty(0, int)
    tracks = tracker1.tracks_c[0]
    upd_for_loose_tracks(tracks, no_targets, 1, tracker1.num_misses_max)
    assert len(tracks) == 1
    assert tracks[0].num_miss == 1

    upd_for_loose_tracks(tracks, no_targets, 1, tracker1.num_misses_max)
    assert len(tracks) == 1
    assert tracks[0].num_miss == 2

    upd_for_loose_tracks(tracks, no_targets, 1, tracker1.num_misses_max)
    assert len(tracks) == 1
    assert tracks[0].num_miss == 3

    upd_for_loose_tracks(tracks, no_targets, 1, tracker1.num_misses_max)
    assert len(tracks) == 0



"""."""

import numpy as np

from kinematic_tracker.tracker.targets import correct_associated


def test_correct_associated() -> None:
    """."""
    r2z = np.ones((1, 6))
    r2id = np.array([567])
    score_rt = np.zeros((1, 0))
    tracks = []
    correct_associated(tracks, np.empty(0, int), np.empty(0, int), score_rt, r2z, r2id)
    assert tracks == []



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.score_copy import ScoreCopy


class Holder:
    def __init__(self) -> None:
        self.cls = -1

    def get_creation_ids_t(self, cls: int) -> np.ndarray:
        self.cls = cls
        return np.linspace(1, 3, 3, dtype=int)


def test_set() -> None:
    sp = ScoreCopy(1)
    score_rt = np.linspace(1, 6, 6).reshape(2, 3)
    holder = Holder()
    sp.set(0, holder.get_creation_ids_t, score_rt)
    assert sp.score_crt[0] == pytest.approx(score_rt)
    assert sp.creation_ids_ct[0] == pytest.approx(holder.get_creation_ids_t(0))
    assert id(sp.score_crt[0]) != id(score_rt)



"""."""

import pytest

from kinematic_tracker import NdKkfTracker
from kinematic_tracker.association.match_hungarian import MatchDriverHungarian
from kinematic_tracker.association.metric_mahalanobis import MetricMahalanobis
from kinematic_tracker.tracker.score_copy import StubScoreCopy


def test_no_measurement_orders() -> None:
    tracker = NdKkfTracker([3, 1], [3, 3])
    assert len(tracker.tracks_c) == 1
    assert tracker.tracks_c[0] == []
    assert isinstance(tracker.score_copy, StubScoreCopy)
    assert isinstance(tracker.metric_driver, MetricMahalanobis)
    assert isinstance(tracker.match_driver, MatchDriverHungarian)
    assert tracker.gen_xz.orders_z == pytest.approx([1, 1])
    assert tracker.gen_xz.gen_x.num_x == 12
    assert tracker.gen_xz.num_z == 6
    assert tracker.buf_rz.shape == (100, 6)

    assert tracker.match_driver.num_reports_max == 100
    assert tracker.match_driver.num_targets_max == 500
    assert tracker.metric_driver.num_reports_max == 100
    assert tracker.metric_driver.num_targets_max == 500



"""."""

from kinematic_tracker import NdKkfTracker
from kinematic_tracker.tracker.score_copy import ScoreCopy


def test_with_return_score() -> None:
    tracker = NdKkfTracker([3, 1], [3, 3], return_score=True)
    assert isinstance(tracker.score_copy, ScoreCopy)



"""."""

from kinematic_tracker import NdKkfTracker
from kinematic_tracker.association.association_metric import AssociationMetric
from kinematic_tracker.association.metric_mahalanobis_size_mod import MetricMahalanobisSizeMod


def test_switch_metric() -> None:
    """."""
    tracker = NdKkfTracker([2, 2], [2, 2])
    name = AssociationMetric.MAHALANOBIS_SIZE_MODULATED.value
    tracker.set_association_metric(name, ind_pos_size=(0, 1, 2, 3))
    assert isinstance(tracker.metric_driver, MetricMahalanobisSizeMod)



from typing import Sequence

from kinematic_tracker.core.track_free_id import get_free_id

from .track import NdKkfTrack


def promote_clutter_eventually(tracks: Sequence[NdKkfTrack], num_det_min: int) -> None:
    """Promote clutter tracks to regular tracks if they meet the
    condition on the minimal number of consecutive associations.
    """
    for track in tracks:
        if track.id < 0 and track.num_det > num_det_min:
            track.id = get_free_id(tracks)



import cv2
import numpy as np
import pytest

from kinematic_tracker.association.metric_giou_yaw import MetricGIoUYaw
from kinematic_tracker.types import FMT


def test_it(
    driver: MetricGIoUYaw, filters8: list[cv2.KalmanFilter], vecs_28z: np.ndarray, ref_88: FMT
) -> None:
    metric_rt = driver.compute_metric(vecs_28z[0], filters8)
    assert metric_rt == pytest.approx(ref_88)



"""."""

import numpy as np

from kinematic_tracker import NdKkfTracker


def test_upd_for_loose_reports1(tracker1: NdKkfTracker) -> None:
    """."""
    o2z = (0.1 + np.linspace(1.0, 6.0, num=6)).reshape(1, 6)
    o2id = np.array([124])
    tracker1.upd_for_loose_reports(tracker1.tracks_c[0], np.zeros(1, int), o2z, o2id)
    assert len(tracker1.tracks_c[0]) == 1
    assert tracker1.tracks_c[0][0].ann_id == 123



"""."""

from kinematic_tracker.association.metric_driver_base import MetricDriverBase


def test_metric_driver_base_init(driver: MetricDriverBase) -> None:
    assert driver.metric_rt.shape == (4, 5)
    assert driver.vec_z.shape == (6, 1)



import numpy as np
import pytest

from kinematic_tracker.association.metric_giou_yaw import MetricGIoUYaw

from .conftest import get_kf_cp


REF_VALUE = 0.5625


def test_normal_1x1_and_2x2(driver: MetricGIoUYaw) -> None:
    det1_rz = np.array([(0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0)])
    kf_cp = get_kf_cp((0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 2.0, 2.0, 2.0))
    metric_rt = driver.compute_metric(det1_rz, [kf_cp])
    assert metric_rt[0, 0] == pytest.approx(REF_VALUE)
    det2_rz = np.array([(0.0, 0.5, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0)])
    metric_rt = driver.compute_metric(det2_rz, [kf_cp])
    assert metric_rt[0, 0] == pytest.approx(REF_VALUE)


def test_normal_2x2_and_1x1(driver: MetricGIoUYaw) -> None:
    det1_rz = np.array([(0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 2.0, 2.0, 2.0)])
    kf_cp = get_kf_cp((0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0))
    metric_rt = driver.compute_metric(det1_rz, [kf_cp])
    assert metric_rt[0, 0] == pytest.approx(REF_VALUE)
    det2_rz = np.array([(0.0, 0.5, 0.0, 1.0, 0.0, 0.0, 1.0, 2.0, 2.0, 2.0)])
    metric_rt = driver.compute_metric(det2_rz, [kf_cp])
    assert metric_rt[0, 0] == pytest.approx(REF_VALUE)



"""."""

import numpy as np
import pytest

from kinematic_tracker.geometry.rectangle_buffers import RectangleBuffers
from kinematic_tracker.types import FMT


@pytest.fixture
def rb5() -> RectangleBuffers:
    return RectangleBuffers(5)


@pytest.fixture
def vec_2z() -> FMT:
    vec_rz = np.linspace(1, 20, 20).reshape(2, 10)
    vec_rz[:, 4:6] = 0
    return vec_rz



"""."""

from kinematic_tracker import NdKkfTracker


def test_with_num_cls() -> None:
    tracker = NdKkfTracker([3, 1], [3, 3], num_cls=12)
    assert len(tracker.tracks_c) == 12
    for tracks in tracker.tracks_c:
        assert tracks == []



"""."""

import numpy as np
import pytest

from kinematic_tracker import NdKkfTracker
from kinematic_tracker.tracker.score_copy import ScoreCopy


def test_with_score_copy(tracker1: NdKkfTracker) -> None:
    tracker1.score_copy = ScoreCopy(1)
    rv = tracker1.advance(2000_000_000, np.linspace(1.1, 6.1, num=6).reshape(1, 6))
    assert isinstance(rv, ScoreCopy)
    assert id(tracker1.score_copy) != id(rv)
    assert rv.score_crt[0] == pytest.approx(0.84785213)
    assert rv.score_crt[0].ndim == 2
    assert rv.creation_ids_ct[0] == pytest.approx([0])



"""The module implements a reference GIOU association metric."""


def giou_lr_lr(seg1: tuple[float, float], seg2: tuple[float, float]) -> float:
    s1, e1 = seg1
    s2, e2 = seg2
    inter_s = max(s1, s2)
    inter_e = min(e1, e2)

    intersection = max(0.0, inter_e - inter_s)
    l1, l2 = e1 - s1, e2 - s2
    union = l1 + l2 - intersection
    if union <= 0:
        return 0.0

    iou = intersection / union

    c_s = min(s1, s2)
    c_e = max(e1, e2)
    c_len = max(0.0, c_e - c_s)

    assert c_len > 0

    giou = iou - (c_len - union) / c_len
    return float((giou + 1) / 2.0)


def giou_ps_ps(ps1: tuple[float, float], ps2: tuple[float, float]) -> float:
    hs1 = ps1[1] / 2
    hs2 = ps2[1] / 2
    seg1 = ps1[0] - hs1, ps1[0] + hs1
    seg2 = ps2[0] - hs2, ps2[0] + hs2
    return giou_lr_lr(seg1, seg2)



"""."""

import numpy as np
import pytest

from kinematic_tracker import NdKkfTracker


@pytest.fixture
def tracker() -> NdKkfTracker:
    """."""
    tracker = NdKkfTracker([3, 1], [3, 3])
    tracker.association.threshold = 0.56
    tracker.set_association_metric('mahalanobis', 0.22 * 6)
    tracker.set_association_method('greedy')
    tracker.num_det_min = 2
    return tracker


@pytest.fixture
def tracker1(tracker: NdKkfTracker) -> NdKkfTracker:
    """."""
    tracker.set_measurement_cov(0.01 * np.eye(6))
    r2z = np.linspace(1.0, 6.0, num=6).reshape(1, 6)
    r2id = np.array([123])
    tracker.advance(1234_000_000, r2z, r2id)
    return tracker


@pytest.fixture
def tc2() -> NdKkfTracker:
    tracker = NdKkfTracker([3, 1], [3, 3], num_cls=2)
    tracker.set_measurement_cov(0.25 * np.eye(6))
    r2z = np.linspace(1.0, 18.0, num=18).reshape(3, 6)
    r2id = np.array([123, 234, 345])
    tracker.advance(1234_000_000, r2z, r2id, cls_r=np.array([0, 1, 1]))
    return tracker



import numpy as np
import pytest

from kinematic_tracker import NdKkfTracker


def test_get_creation_ids_one(tracker1: NdKkfTracker) -> None:
    ids = tracker1.get_creation_ids(0)
    assert ids == pytest.approx([0])


def test_get_creation_ids_two(tracker1: NdKkfTracker) -> None:
    det_rz = np.linspace(1.0, 12, 12).reshape(2, 6)
    tracker1.advance(2000_000_000, det_rz)
    ids = tracker1.get_creation_ids(0)
    assert ids == pytest.approx([0, 1])


def test_multi_class(tc2: NdKkfTracker) -> None:
    ids0 = tc2.get_creation_ids(0)
    assert ids0.shape == (1,)
    assert ids0.tolist() == [0]

    ids1 = tc2.get_creation_ids(1)
    assert ids1.shape == (2,)
    assert ids1.tolist() == [1, 2]



"""."""

import numpy as np

from kinematic_tracker import NdKkfTracker


def test_zero_detections_default(tracker: NdKkfTracker) -> None:
    tracker.set_measurement_cov(0.01 * np.eye(6))
    tracker.advance(100_000_000, np.empty((0, 6)))
    assert len(tracker.tracks_c[0]) == 0



"""."""

from kinematic_tracker.tracker.score_copy import ScoreCopy, StubScoreCopy, get_score_copy


def test_get_score_copy() -> None:
    assert isinstance(get_score_copy(False, 1), StubScoreCopy)
    assert isinstance(get_score_copy(True, 1), ScoreCopy)



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.score_copy import ScoreCopy


def test_deep_copy() -> None:
    sp = ScoreCopy(1)
    sp.score_crt[0] = np.linspace(1, 6, 6).reshape(2, 3)
    sp.creation_ids_ct[0] = np.linspace(1, 3, 3, dtype=int)
    sp_cp = sp.deep_copy()
    assert isinstance(sp_cp, ScoreCopy)
    assert id(sp_cp) != id(sp)
    assert sp_cp.score_crt[0] == pytest.approx(sp.score_crt[0])
    assert sp_cp.creation_ids_ct[0] == pytest.approx(sp.creation_ids_ct[0])
    assert id(sp_cp.score_crt[0]) != id(sp.score_crt[0])
    assert id(sp_cp.creation_ids_ct[0]) != id(sp.creation_ids_ct[0])



"""."""

import pytest

from kinematic_tracker.association.metric_giou_aligned import MetricGIoUAligned


@pytest.fixture
def driver() -> MetricGIoUAligned:
    return MetricGIoUAligned(100, 500, 6, (0, 1, 2, 3, 4, 5))



"""."""

import numpy as np
import pytest

from kinematic_tracker.geometry.giou_yaw_aux import GiouYawAux
from kinematic_tracker.geometry.rectangle_buffers import RectangleBuffers


def test_compute(aux: GiouYawAux) -> None:
    # prepare
    rect_buffers = RectangleBuffers(10)
    det_rz = np.linspace(1, 40, 40).reshape(4, 10)
    det_rz[:, 4:6] = 0
    rect_buffers.set_nu_scenes(det_rz)
    cuboid = det_rz[1, :]
    poly = rect_buffers.mid_polygons_n[1]
    polygons = rect_buffers.mid_polygons_n[:4]
    volumes = det_rz[:, -3:].prod(axis=-1)
    out = np.ones(4)
    aux.set_first_4_vertices(rect_buffers.rect_coordinates_n4c[1, :, :2], 4)
    aux.set_second_4_vertices(rect_buffers.rect_coordinates_n4c[:4, :, :2])
    # compute
    aux.compute(cuboid, poly, det_rz, polygons, volumes, out)
    ref = 0.3185216546083782, 1.0, 0.4392318438129806, 0.3261591198359483
    assert out == pytest.approx(ref)



import numpy as np
import pytest

from kinematic_tracker.tracker.check import check_cls_r


def test_pass_with_zero_size() -> None:
    check_cls_r(np.zeros(0, int), 0, 4)


def test_when_sizes_are_wrong() -> None:
    cls_r = np.linspace(0, 3, 4, dtype=int)
    with pytest.raises(ValueError) as err_size:
        check_cls_r(cls_r, 5, 4)
    assert err_size.value.args[0] == 'Number of reports should be 5, but got 4.'

    with pytest.raises(ValueError) as err_max:
        check_cls_r(cls_r, 4, 3)
    assert err_max.value.args[0] == 'Class IDs should be less than 3, but I see 3.'


def test_when_negative_classes() -> None:
    cls_r = np.linspace(-4, 3, 8, dtype=int)
    with pytest.raises(ValueError) as err:
        check_cls_r(cls_r, 8, 4)
    assert err.value.args[0] == 'Class IDs should be greater or equal zero, but I see -4.'



"""."""

from kinematic_tracker import NdKkfTracker
from kinematic_tracker.tracker.promote_clutter import promote_clutter_eventually


def test_promote_clutter() -> None:
    tracks = []
    promote_clutter_eventually(tracks, 3)
    assert len(tracks) == 0


def test_promote_clutter1(tracker1: NdKkfTracker) -> None:
    tracker1.tracks_c[0][0].num_det = 5
    promote_clutter_eventually(tracker1.tracks_c[0], tracker1.num_det_min)
    assert tracker1.tracks_c[0][0].id == 0



"""."""

import pytest

from kinematic_tracker.core.giou_1d import giou_ps_ps


def test_full_overlap() -> None:
    assert giou_ps_ps((0, 2), (0, 2)) == pytest.approx(1.0)


def test_symmetrical() -> None:
    assert giou_ps_ps((0, 2), (0, 4)) == pytest.approx(0.75)
    assert giou_ps_ps((0, 4), (0, 2)) == pytest.approx(0.75)


def test_zero_len_non_diff_loc() -> None:
    assert giou_ps_ps((1, 0), (2, 0)) == pytest.approx(0.0)


def test_no_overlap_unit_len() -> None:
    assert giou_ps_ps((1.5, 1.0), (3.5, 1.0)) == pytest.approx(0.3333333333333333)


def test_no_overlap_len_1_0() -> None:
    assert giou_ps_ps((1.5, 1), (3, 0)) == pytest.approx(0.25)
